// Standard Libraries
#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <stdatomic.h>

// Common Libraries
#include <string.h>
#include <pthread.h>
#include <inttypes.h>

// CLI and Hashing
#include <argparse.h>
#include <xxhash.h>

// Strings
#define PICO_STRING_IMPLEMENTATION
#include <pico_string.h>

// CSV
#define PICO_CSV_IMPLEMENTATION
#include <pico_csv.h>

// ---------------------------- Definitions

#define CACHE_LINE_SIZE 16
#define BASE_ALLOC_SIZE 1024

// ---------------------------- IntArray

typedef struct IntArray
{
	uint32_t *data;
	uint32_t size;
}
IntArray;

IntArray *IntArray_from_string (char *string);
void      IntArray_delete      (IntArray *array);
void      IntArray_print       (IntArray *array);
bool      IntArray_contains    (IntArray *array, uint32_t value);
IntArray *IntArray_distribute  (uint64_t size, uint32_t num_workers);

// ---------------------------- StrArray

typedef struct StrArray
{
	pstr *data;
	uint64_t size;
	uint64_t alloc_size;
}
StrArray;

StrArray *StrArray_new    (uint64_t alloc_size);
void      StrArray_delete (StrArray *array);
void      StrArray_adjust (StrArray *str_array);
void      StrArray_append (StrArray *str_array, pstr string);

// ---------------------------- HashTable

typedef enum BucketState
{
	BUCKET_EMPTY,
	BUCKET_WRITE,
	BUCKET_DONE
}
BucketState;

typedef struct Bucket
{
	uint64_t hash;
	BucketState state;
}
Bucket;

typedef _Atomic(Bucket)
AtomicBucket;

typedef struct Data
{
	pstr key;
	uint64_t value;
	uint64_t key_row;
}
Data;

typedef struct HashTable
{
	AtomicBucket *buckets;
	Data *data;
	uint64_t size;
}
HashTable;

HashTable *HashTable_new    (uint64_t size);
void       HashTable_delete (HashTable *table);
bool       HashTable_insert (HashTable *table, Data data);
bool       HashTable_get    (HashTable *table, Data *data);

// ---------------------------- TableData

typedef struct TableData
{
	StrArray *rows;
	StrArray *keys;
	HashTable *table;
}
TableData;

TableData *TableData_from_file (const char *filename);
void       TableData_delete    (TableData *table_data);

// ---------------------------- Thread Arguments

typedef struct ProcessCsvRowsArgs
{
	StrArray *rows;
	StrArray *keys;
	uint64_t row_count;
	uint64_t row_offset;

	HashTable *table;
	IntArray *primary_keys;
	IntArray *ignore_columns;
}
ProcessCsvRowsArgs;

typedef struct FindCsvInsertsArgs
{
	StrArray *inserts;
	atomic_bool *done;

	TableData *base;
	TableData *delta;
	uint64_t delta_row_count;
	uint64_t delta_row_offset;
}
FindCsvInsertsArgs;

typedef struct FindCsvDeletesUpdatesArgs
{
	StrArray *deletes;
	StrArray *updates;
	atomic_bool *done;

	TableData *base;
	TableData *delta;
	uint64_t base_row_count;
	uint64_t base_row_offset;
}
FindCsvDeletesUpdatesArgs;

void *process_csv_rows         (void *args_ptr);
void *find_csv_inserts         (void *args_ptr);
void *find_csv_deletes_updates (void *args_ptr);

// ---------------------------- Environment

typedef struct Environment
{
	// Global error
	atomic_bool error;

	// Boolean flags
	bool debug_mode;
	bool show_count;
	bool show_changes;
	bool only_updates;

	// Environment constants
	bool disable_free;
	double load_factor;
	uint32_t thread_count;

	// CSV settings
	IntArray *primary_keys;
	IntArray *ignore_columns;
	char *raw_primary_keys;
	char *raw_ignore_columns;

	// File names
	const char *base_filename;
	const char *delta_filename;
	char *insert_filename;
	char *delete_filename;
	char *update_filename;
}
Environment;

static Environment env;

static const char *usages[] =
{
	"csvdiff [options] [--] <base.csv> <delta.csv>",
	NULL,
};

static struct argparse_option options[] =
{
	OPT_HELP(),

	OPT_GROUP("Boolean flags"),
	OPT_BOOLEAN('d', "debug-mode",   &env.debug_mode,   "enable debug mode",          NULL, 0, 0),
	OPT_BOOLEAN('c', "show-count",   &env.show_count,   "show the number of changes", NULL, 0, 0),
	OPT_BOOLEAN('s', "show-changes", &env.show_changes, "show all file changes",      NULL, 0, 0),
	OPT_BOOLEAN('u', "only-updates", &env.only_updates, "only save updated rows",     NULL, 0, 0),

	OPT_GROUP("Environment constants"),
	OPT_BOOLEAN('f', "disable-free", &env.disable_free, "disable final calls of free()",  NULL, 0, 0),
	OPT_FLOAT  ('l', "load-factor",  &env.load_factor,  "load factor for the hash table", NULL, 0, 0),
	OPT_INTEGER('t', "thread-count", &env.thread_count, "number of threads to use",       NULL, 0, 0),

	OPT_GROUP("CSV settings"),
	OPT_STRING ('p', "primary-keys",   &env.raw_primary_keys,   "list of primary keys",      NULL, 0, 0),
	OPT_STRING ('i', "ignore-columns", &env.raw_ignore_columns, "list of columns to ignore", NULL, 0, 0),

	OPT_GROUP("File names"),
	OPT_STRING('I', "inserts", &env.insert_filename, "inserts CSV file name, can be NULL", NULL, 0, 0),
	OPT_STRING('D', "deletes", &env.delete_filename, "deletes CSV file name, can be NULL", NULL, 0, 0),
	OPT_STRING('U', "updates", &env.update_filename, "updates CSV file name, can be NULL", NULL, 0, 0),

	OPT_END()
};

void Environment_init (void);
void Environment_exit (void);

// ---------------------------- IntArray

IntArray *IntArray_from_string (char *string)
{
	IntArray *list = malloc(sizeof(IntArray));
	pstr str_copy = pstr_new(string);
	char *token = strtok(str_copy, ",");

	list->data = NULL;
	list->size = 0;

	while (token)
	{
		list->data = realloc(list->data, (list->size + 1) * sizeof(uint32_t));
		list->data[list->size++] = strtoul(token, NULL, 10);
		token = strtok(NULL, ",");
	}

	pstr_delete(str_copy);
	return list;
}

void IntArray_delete (IntArray *array)
{
	free(array->data);
	free(array);
}

void IntArray_print (IntArray *array)
{
	for (uint32_t i = 0; i < array->size; i++)
	{
		printf("%"PRIu32, array->data[i]);
		if (i < array->size - 1)
			printf(", ");
	}
}

bool IntArray_contains (IntArray *array, uint32_t value)
{
	for (uint64_t i = 0; i < array->size; i++)
		if (array->data[i] == value)
			return true;

	return false;
}

IntArray *IntArray_distribute (uint64_t size, uint32_t num_workers)
{
	IntArray *array = malloc(sizeof(IntArray));

	if (size < num_workers)
		num_workers = size;

	array->data = malloc(num_workers * sizeof(uint32_t));
	array->size = num_workers;

	uint32_t division = size / num_workers;
	uint32_t remainder = size % num_workers;

	for (uint32_t i = 0; i < num_workers - 1; i++)
		array->data[i] = division;

	array->data[num_workers - 1] = division + remainder;
	return array;
}

// ---------------------------- StrArray

StrArray *StrArray_new (uint64_t alloc_size)
{
	StrArray *array = malloc(sizeof(StrArray));
	array->data = malloc(alloc_size * sizeof(pstr));
	array->alloc_size = alloc_size;
	array->size = 0;
	return array;
}

void StrArray_delete (StrArray *array)
{
	free(array->data);
	free(array);
}

void StrArray_adjust (StrArray *str_array)
{
	str_array->data = realloc(str_array->data, str_array->size * sizeof(pstr));
	str_array->alloc_size = str_array->size;
}

void StrArray_append (StrArray *str_array, pstr string)
{
	if (str_array->size >= str_array->alloc_size)
	{
		str_array->alloc_size = str_array->size / env.load_factor;
		if (str_array->alloc_size == str_array->size) str_array->alloc_size++;
		str_array->data = realloc(str_array->data, str_array->alloc_size * sizeof(pstr));
	}

	str_array->data[str_array->size] = string;
	str_array->size++;
}

// ---------------------------- HashTable

HashTable *HashTable_new (uint64_t size)
{
	HashTable *table = malloc(sizeof(HashTable));

	// Adjusting the size with the load factor
	uint64_t real_size = size / env.load_factor;
	if (real_size == size) real_size++;

	table->buckets = calloc(real_size, sizeof(AtomicBucket));
	table->data = calloc(real_size, sizeof(Data));
	table->size = real_size;
	return table;
}

void HashTable_delete (HashTable *table)
{
	free(table->buckets);
	free(table->data);
	free(table);
}

// Return true if the key is already in the table, and false otherwise
bool HashTable_insert (HashTable *table, Data data)
{
	uint64_t key_len = pstr_info(data.key).length;
	uint64_t hash = XXH3_64bits(data.key, key_len);
	uint64_t index = hash % table->size;
	uint64_t hash_seed = 0;
	Bucket bucket;

	while (true)
	{
		for (uint64_t cache_index = 0; cache_index < CACHE_LINE_SIZE; cache_index++)
		{
			Bucket bucket_empty = {0, BUCKET_EMPTY};
			Bucket bucket_done = {hash, BUCKET_DONE};
			Bucket bucket_write = {hash, BUCKET_WRITE};
			uint64_t i = (index + cache_index) % table->size;

			// Bucket with EMPTY state is found
			if ((bucket = atomic_load(&table->buckets[i])).state == BUCKET_EMPTY)
			{
				// The bucket is set to WRITE state if it matches the old state
				if (atomic_compare_exchange_strong(&table->buckets[i], &bucket_empty, bucket_write))
				{
					// The data is written to the table
					table->data[i] = data;

					// The bucket is set to DONE state
					atomic_store(&table->buckets[i], bucket_done);

					// The key was not in the table
					return false;
				}

				// The bucket state is equal to the old state, but the CAS failed
				else if ((bucket = atomic_load(&table->buckets[i])).state == BUCKET_EMPTY)
				{
					if (env.debug_mode)
						printf("WARNING: HashTable_insert() -> Strong CAS failed\n");
					cache_index--;
				}
			}

			// Bucket with DONE state and matching hash is found
			else if ((bucket = atomic_load(&table->buckets[i])).hash == hash)
			{
				// Waiting for the bucket to exit the WRITE state
				while ((bucket = atomic_load(&table->buckets[i])).state == BUCKET_WRITE);

				// The key also matches
				if (pstr_pscmp(table->data[i].key, data.key) == 0)
				{
					// The data is updated
					atomic_store(&table->buckets[i], bucket_write);
					table->data[i] = data;
					atomic_store(&table->buckets[i], bucket_done);

					// The key was already in the table
					return true;
				}
			}
		}

		index = XXH3_64bits_withSeed(data.key, key_len, ++hash_seed) % table->size;
	}

	return false;
}

// Return true if the key is found in the table, and false otherwise
bool HashTable_get (HashTable *table, Data *data)
{
	uint64_t key_len = pstr_info(data->key).length;
	uint64_t hash = XXH3_64bits(data->key, key_len);
	uint64_t index = hash % table->size;
	uint64_t hash_seed = 0;
	Bucket bucket;

	while (true)
	{
		for (uint64_t cache_index = 0; cache_index < CACHE_LINE_SIZE; cache_index++)
		{
			uint64_t i = (index + cache_index) % table->size;

			// Bucket with EMPTY state is found
			if ((bucket = atomic_load(&table->buckets[i])).state == BUCKET_EMPTY)
				return false;

			// Bucket with DONE state and matching hash is found
			else if ((bucket = atomic_load(&table->buckets[i])).hash == hash)
			{
				// Waiting for the bucket to exit the WRITE state
				while ((bucket = atomic_load(&table->buckets[i])).state == BUCKET_WRITE);

				// The key is already in the table
				if (pstr_pscmp(table->data[i].key, data->key) == 0)
				{
					*data = table->data[i];
					return true;
				}
			}
		}

		index = XXH3_64bits_withSeed(data->key, key_len, ++hash_seed) % table->size;
	}

	return false;
}

// ---------------------------- TableData

TableData *TableData_from_file (const char *filename)
{
	// Creating a new CSV reader
	CsvReader *csv = CsvReader_new(filename);

	// Verifying that the reader was created successfully
	if (csv == NULL)
	{
		if (env.debug_mode)
			printf("ERROR: Failed to read CSV file '%s'\n", filename);
		env.error = true;
		return NULL;
	}

	// Initializing the rows buffer
	StrArray *rows = StrArray_new(BASE_ALLOC_SIZE);
	char *row;

	if (env.debug_mode)
		printf("Loading CSV file into memory...\n");

	// Read all rows into memory
	while ((row = CsvReader_next_row(csv)))
	{
		// Obtaining the row length and appending it into the 'rows' array
		uint64_t row_len = CsvReader_row_info(csv).raw_len;
		StrArray_append(rows, pstr_new_exp(row, row_len));
	}

	// Adjusting the 'rows' array to fit its contents
	StrArray_adjust(rows);

	if (env.debug_mode)
		printf("Done! [%"PRIu64" rows]\n", rows->size);

	// Deleting the CSV reader
	CsvReader_delete(csv);

	// Creating the 'keys' array and the hash table
	StrArray *keys = StrArray_new(rows->size);
	HashTable *table = HashTable_new(rows->size);
	keys->size = rows->size;

	// Declaring the threads and the arguments
	pthread_t threads[env.thread_count];
	ProcessCsvRowsArgs args[env.thread_count];

	// Creating a work distribution for the threads
	IntArray *work_distrib = IntArray_distribute(rows->size, env.thread_count);

	// Declaring the base thread arguments
	ProcessCsvRowsArgs common_args =
	{
		.rows = rows,
		.keys = keys,
		.table = table,
		.primary_keys = env.primary_keys,
		.ignore_columns = env.ignore_columns
	};

	if (env.debug_mode)
		printf("Loading data into hash table...\n");

	// Creating the threads
	for (uint32_t i = 0; i < work_distrib->size; i++)
	{
		args[i] = common_args;
		args[i].row_count = work_distrib->data[i];
		args[i].row_offset = work_distrib->data[0] * i;
		pthread_create(&threads[i], NULL, process_csv_rows, &args[i]);
	}

	// Joining the threads
	for (uint32_t i = 0; i < work_distrib->size; i++)
		pthread_join(threads[i], NULL);

	if (env.debug_mode)
		printf("Done!\n\n");

	// Deleting the work distribution
	IntArray_delete(work_distrib);

	// Creating a new table data object and assigning the values
	TableData *table_data = malloc(sizeof(TableData));
	*table_data = (TableData) {rows, keys, table};

	return table_data;
}

void TableData_delete (TableData *table_data)
{
	for (register uint64_t i = 0; i < table_data->rows->size; i++)
	{
		pstr_delete(table_data->rows->data[i]);
		pstr_delete(table_data->keys->data[i]);
	}

	StrArray_delete(table_data->rows);
	StrArray_delete(table_data->keys);
	HashTable_delete(table_data->table);
	free(table_data);
}

// ---------------------------- Thread Functions

// Processes the rows of a CSV file and loads them into a hash table
void *process_csv_rows (void *args_ptr)
{
	ProcessCsvRowsArgs *args = args_ptr;
	CsvReader *col_context = CsvReader_new(NULL);
	XXH3_state_t *hash_state = XXH3_createState();
	pstr row_copy = pstr_new_exp(NULL, BASE_ALLOC_SIZE);

	for (register uint64_t i = 0; i < args->row_count; i++)
	{
		uint64_t col_index = 0, key_row = args->row_offset + i;
		pstr row = args->rows->data[key_row];
		pstr key = pstr_new(NULL);
		char *col;

		row_copy = pstr_pscpy(row_copy, row);
		XXH3_64bits_reset(hash_state);

		while ((col = CsvReader_next_col(col_context, row_copy)))
		{
			uint64_t col_len = CsvReader_col_info(col_context).parse_len;

			if (IntArray_contains(args->primary_keys, col_index))
				key = pstr_cat_exp(key, col, col_len);
			else if (!IntArray_contains(args->ignore_columns, col_index))
				XXH3_64bits_update(hash_state, col, col_len);

			col_index++;
		}

		uint64_t value = XXH3_64bits_digest(hash_state);
		Data row_data = {key, value, key_row};

		if (HashTable_insert(args->table, row_data) && env.debug_mode)
			printf("WARNING: Duplicate key '%s' found in row %"PRIu64"\n", key, key_row);

		args->keys->data[key_row] = key;
	}

	pstr_delete(row_copy);
	XXH3_freeState(hash_state);
	CsvReader_delete(col_context);
	return NULL;
}

// Finds the rows that were inserted in the delta CSV file
void *find_csv_inserts (void *args_ptr)
{
	FindCsvInsertsArgs *args = args_ptr;

	for (register uint64_t i = 0; i < args->delta_row_count; i++)
	{
		uint64_t delta_key_row = args->delta_row_offset + i;
		pstr delta_key = args->delta->keys->data[delta_key_row];
		Data delta_data = {.key = delta_key}, base_data = {.key = delta_key};

		// Base table does not contain delta key
		if (!HashTable_get(args->base->table, &base_data))
		{
			// Delta data is retrieved from the delta table
			HashTable_get(args->delta->table, &delta_data);

			// Only rows that have not been processed are inserted
			if (atomic_compare_exchange_strong(&args->done[delta_data.key_row], &(bool) {false}, true))
			{
				pstr delta_row = args->delta->rows->data[delta_data.key_row];
				StrArray_append(args->inserts, delta_row);
			}

			// The row is not done, but the strong CAS failed
			else if (args->done[delta_data.key_row] == false)
			{
				if (env.debug_mode)
					printf("WARNING: find_csv_inserts() -> Strong CAS failed\n");
				i--;
			}
		}
	}

	return NULL;
}

// Finds the rows that were deleted and updated in the delta CSV file
void *find_csv_deletes_updates (void *args_ptr)
{
	FindCsvDeletesUpdatesArgs *args = args_ptr;

	for (register uint64_t i = 0; i < args->base_row_count; i++)
	{
		uint64_t base_key_row = args->base_row_offset + i;
		pstr base_key = args->base->keys->data[base_key_row];
		Data base_data = {.key = base_key}, delta_data = {.key = base_key};

		// Base data is retrieved from the base table
		HashTable_get(args->base->table, &base_data);

		// Only rows that have not been processed are deleted or updated
		if (atomic_compare_exchange_strong(&args->done[base_data.key_row], &(bool) {false}, true))
		{
			// Base key is found in the delta table
			if (HashTable_get(args->delta->table, &delta_data))
			{
				// If the values are different, the row was updated
				if (base_data.value != delta_data.value)
				{
					pstr base_row = args->base->rows->data[base_data.key_row];
					pstr delta_row = args->delta->rows->data[delta_data.key_row];

					if (env.only_updates == false)
						StrArray_append(args->updates, base_row);

					StrArray_append(args->updates, delta_row);
				}
			}

			// Base key is not found in the delta table, so the row was deleted
			else
			{
				pstr base_row = args->base->rows->data[base_data.key_row];
				StrArray_append(args->deletes, base_row);
			}
		}

		// The row is not done, but the strong CAS failed
		else if (args->done[delta_data.key_row] == false)
		{
			if (env.debug_mode)
				printf("WARNING: find_csv_deletes_updates() -> Strong CAS failed\n");
			i--;
		}
	}

	return NULL;
}

// ---------------------------- Environment

void Environment_init (void)
{
	// Boolean flags
	env.error = false;
	env.debug_mode = false;
	env.show_count = false;
	env.show_changes = false;
	env.only_updates = false;

	// Environment constants
	env.disable_free = false;
	env.load_factor = 0.75;
	env.thread_count = 1;

	// CSV settings
	env.primary_keys = NULL;
	env.ignore_columns = NULL;
	env.raw_primary_keys = "0";
	env.raw_ignore_columns = "";

	// File names
	env.base_filename = NULL;
	env.delta_filename = NULL;
	env.insert_filename = "insert.csv";
	env.delete_filename = "delete.csv";
	env.update_filename = "update.csv";
}

void Environment_exit (void)
{
	IntArray_delete(env.ignore_columns);
	IntArray_delete(env.primary_keys);
}

// ---------------------------- Main Functions

// Finds the rows that were inserted in the delta CSV file
void find_inserts (FindCsvInsertsArgs *args, IntArray *work, TableData *base, TableData *delta)
{
	// Debug message
	if (env.debug_mode)
		printf("Finding inserts...\n");

	pthread_t threads[env.thread_count];
	atomic_bool *done = calloc(delta->rows->size, sizeof(atomic_bool));
	FindCsvInsertsArgs common_args = {.base = base, .delta = delta, .done = done};

	// Generating the threads to find the inserts
	for (uint32_t i = 0; i < work->size; i++)
	{
		args[i] = common_args;
		args[i].inserts = StrArray_new(BASE_ALLOC_SIZE);
		args[i].delta_row_count = work->data[i];
		args[i].delta_row_offset = work->data[0] * i;
		pthread_create(&threads[i], NULL, find_csv_inserts, &args[i]);
	}

	// Finding the inserts
	for (uint32_t i = 0; i < work->size; i++)
		pthread_join(threads[i], NULL);

	free(done);

	// Debug message
	if (env.debug_mode)
		printf("Done!\n\n");
}

// Finds the rows that were deleted or updated in the delta CSV file
void find_deletes_updates (FindCsvDeletesUpdatesArgs *args, IntArray *work, TableData *base, TableData *delta)
{
	// Debug message
	if (env.debug_mode)
		printf("Finding deletes and updates...\n");

	pthread_t threads[env.thread_count];
	atomic_bool *done = calloc(base->rows->size, sizeof(atomic_bool));
	FindCsvDeletesUpdatesArgs common_args = {.base = base, .delta = delta, .done = done};

	// Generating the threads to find the deletes and updates
	for (uint32_t i = 0; i < work->size; i++)
	{
		args[i] = common_args;
		args[i].deletes = StrArray_new(BASE_ALLOC_SIZE);
		args[i].updates = StrArray_new(BASE_ALLOC_SIZE);
		args[i].base_row_count = work->data[i];
		args[i].base_row_offset = work->data[0] * i;
		pthread_create(&threads[i], NULL, find_csv_deletes_updates, &args[i]);
	}

	// Finding the deletes and updates
	for (uint32_t i = 0; i < work->size; i++)
		pthread_join(threads[i], NULL);

	free(done);

	// Debug message
	if (env.debug_mode)
		printf("Done!\n\n");
}

void write_file (char *file_name, StrArray *content[], uint64_t count, pstr header)
{
	if (strcmp(file_name, "NULL") == 0)
		return;

	FILE *output = fopen(file_name, "wb");

	// Debug message if the file could not be opened
	if (env.debug_mode && output == NULL)
		printf("ERROR: Could not open the file %s for writing!\n", file_name);

	// Only writes to the file if it was opened
	if (output != NULL)
	{
		// Writing the header
		fprintf(output, "%s\n", header);

		// Writing the content
		for (uint64_t i = 0; i < count; i++)
		{
			for (uint64_t j = 0; j < content[i]->size; j++)
				fprintf(output, "%s\n", content[i]->data[j]);
		}

		// Closing the file
		fclose(output);
	}
}

// Finds and saves the differences between the base and delta CSV files
void find_save_difference (TableData *base, TableData *delta)
{
	// Arguments for the threads
	FindCsvInsertsArgs insert_args[env.thread_count];
	FindCsvDeletesUpdatesArgs delete_update_args[env.thread_count];

	// Work distributions
	IntArray *delta_work = IntArray_distribute(delta->rows->size, env.thread_count);
	IntArray *base_work = IntArray_distribute(base->rows->size, env.thread_count);

	// Finding the inserts, deletes and updates
	find_inserts(insert_args, delta_work, base, delta);
	find_deletes_updates(delete_update_args, base_work, base, delta);

	// Counters for the number of inserts, deletes and updates
	uint64_t insert_count = 0;
	uint64_t delete_count = 0;
	uint64_t update_count = 0;

	// 2D arrays of the inserts, deletes and updates
	StrArray *inserts[delta_work->size];
	StrArray *deletes[base_work->size];
	StrArray *updates[base_work->size];

	// Filling the inserts array with the data from the threads
	for (uint32_t i = 0; i < delta_work->size; i++)
	{
		inserts[i] = insert_args[i].inserts;
		insert_count += inserts[i]->size;
	}

	// Filling the deletes and updates arrays with the data from the threads
	for (uint32_t i = 0; i < base_work->size; i++)
	{
		deletes[i] = delete_update_args[i].deletes;
		updates[i] = delete_update_args[i].updates;
		delete_count += deletes[i]->size;
		update_count += updates[i]->size;
	}

	// Writing the inserts, deletes and updates to the output files
	pstr header = base->rows->data[0];
	write_file(env.insert_filename, inserts, delta_work->size, header);
	write_file(env.delete_filename, deletes, base_work->size,  header);
	write_file(env.update_filename, updates, base_work->size,  header);

	// Adjusting the update count if the original row is included in the update
	if (env.only_updates == false)
		update_count /= 2;

	// Printing the changes to the console
	if (env.show_changes)
	{
		printf("# Insertions (%"PRIu64")\n", insert_count);

		for (uint32_t i = 0; i < delta_work->size; i++)
		{
			for (uint64_t j = 0; j < inserts[i]->size; j++)
				printf("+ %s\n", inserts[i]->data[j]);
		}

		printf("\n# Deletions: (%"PRIu64")\n", delete_count);

		for (uint32_t i = 0; i < base_work->size; i++)
		{
			for (uint64_t j = 0; j < deletes[i]->size; j++)
				printf("- %s\n", deletes[i]->data[j]);
		}

		printf("\n# Updates: (%"PRIu64")\n", update_count);

		for (uint32_t i = 0; i < base_work->size; i++)
		{
			for (uint64_t j = 0; j < updates[i]->size; j += env.only_updates ? 1 : 2)
			{
				if (env.only_updates == false)
				{
					printf("- %s\n", updates[i]->data[j]);
					printf("+ %s\n", updates[i]->data[j + 1]);
				}
				else
					printf("* %s\n", updates[i]->data[j]);
			}
		}

		if (env.debug_mode)
			printf("\n");
	}
	// Printing the counts to the console
	else if (env.show_count)
	{
		printf("Insertions: %"PRIu64"\n", insert_count);
		printf("Deletions:  %"PRIu64"\n", delete_count);
		printf("Updates:    %"PRIu64"\n", update_count);
	}

	// Deleting the insert arrays
	for (uint32_t i = 0; i < delta_work->size; i++)
		StrArray_delete(insert_args[i].inserts);

	// Deleting the delete and update arrays
	for (uint32_t i = 0; i < base_work->size; i++)
	{
		StrArray_delete(delete_update_args[i].deletes);
		StrArray_delete(delete_update_args[i].updates);
	}

	// Deleting the work distribution arrays
	IntArray_delete(base_work);
	IntArray_delete(delta_work);
}

int main (int argc, const char **argv)
{
	if (argc == 1)
	{
		argc = 2;
		argv = (const char *[]) { argv[0], "-h" };
		printf("Program called with no arguments...\n\n");
	}

	Environment_init();
	struct argparse argparse;
	argparse_init(&argparse, options, usages, 0);
	argparse_describe(&argparse, "\nFinds the difference between two CSV files", NULL);
	argc = argparse_parse(&argparse, argc, argv);

	// Parsing the environment lists
	env.primary_keys = IntArray_from_string(env.raw_primary_keys);
	env.ignore_columns = IntArray_from_string(env.raw_ignore_columns);

	if (argc == 2)
	{
		env.base_filename = argv[0];
		env.delta_filename = argv[1];
	}
	else
	{
		if (env.debug_mode)
			printf("ERROR: Invalid number of arguments!\n");
		return EXIT_FAILURE;
	}

	// Printing the environment variables
	if (env.debug_mode)
	{
		if (env.disable_free)
			printf("NOTE: Final calls of free() are disabled\n");

		// Output file names
		if (strcmp(env.insert_filename, "NULL") == 0)
			printf("NOTE: Insert file is disabled\n");
		if (strcmp(env.delete_filename, "NULL") == 0)
			printf("NOTE: Delete file is disabled\n");
		if (strcmp(env.update_filename, "NULL") == 0)
			printf("NOTE: Update file is disabled\n");

		// Number of threads to use
		printf("Thread count: %"PRIu32"\n", env.thread_count);

		// Primary keys of the table
		printf("Primary keys: [");
		IntArray_print(env.primary_keys);
		printf("]\n");

		// Ignored columns of the table
		printf("Ignored columns: [");
		IntArray_print(env.ignore_columns);
		printf("]\n\n");
	}

	// Loading the base file
	if (env.debug_mode)
		printf("Loading base file...\n");
	TableData *base = TableData_from_file(env.base_filename);
	if (env.error) return EXIT_FAILURE;

	// Loading the delta file
	if (env.debug_mode)
		printf("Loading delta file...\n");
	TableData *delta = TableData_from_file(env.delta_filename);
	if (env.error) return EXIT_FAILURE;

	// Finding the difference between the two files
	find_save_difference(base, delta);

	// Freeing the table data
	if (env.disable_free == false)
	{
		if (base != NULL)  TableData_delete(base);
		if (delta != NULL) TableData_delete(delta);
	}

	// Printing the result of the program
	if (env.debug_mode)
		printf("Done! [%s]\n", env.error ? "ERROR" : "SUCCESS");

	Environment_exit();
	return env.error ? EXIT_FAILURE : EXIT_SUCCESS;
}