// Pico CSV Reading Library
// by. Foxuper

#ifndef PICO_CSV_H
#define PICO_CSV_H

// Includes
#include <stdint.h>

// The file is scanned in 'CSV_BUFFER_SIZE' chunks
#define CSV_BUFFER_SIZE (64 * 1024 * 1024)

// Row information type
typedef struct CsvRowInfo
{
	uint64_t raw_len;
	uint64_t file_offset;
}
CsvRowInfo;

// Column information type
typedef struct CsvColInfo
{
	uint64_t raw_len;
	uint64_t parse_len;
	uint64_t row_offset;
}
CsvColInfo;

// CSV Reader type (intended to be opaque)
typedef struct CsvReaderPrivate CsvReader;

// CSV Reader functions
extern CsvReader *CsvReader_new      (const char *filename);
extern CsvReader *CsvReader_new_full (const char *filename, char delim_char, char escape_char);
extern void       CsvReader_reset    (CsvReader *csv);
extern void       CsvReader_delete   (CsvReader *csv);
extern char      *CsvReader_next_row (CsvReader *csv);
extern char      *CsvReader_next_col (CsvReader *csv, char *row);
extern CsvRowInfo CsvReader_row_info (CsvReader *csv);
extern CsvColInfo CsvReader_col_info (CsvReader *csv);

#endif // PICO_CSV_H



#ifdef PICO_CSV_IMPLEMENTATION

struct CsvReaderPrivate
{
	FILE *file; // CSV file handle

	char *buffer; // Buffer that holds a portion of the file
	char *row;    // Buffer that holds the current row

	char delim_char;  // Delimiter character (usually ',')
	char escape_char; // Escape character (usually '"')

	char *buffer_end; // Pointer to the end of the buffer
	char *buffer_cur; // Pointer to the current position in the buffer
	char *column_cur; // Pointer to the current column in the row

	char     col_last_char;  // Last character read in the previous column
	uint64_t col_raw_len;    // Length of the current column (as it appears in the file)
	uint64_t col_parse_len;  // Length of the current column (after it has been parsed)
	uint64_t col_row_offset; // Offset of the current column in the row being parsed

	uint64_t row_raw_len;        // Length of the current row (as it appears in the file)
	uint64_t row_buffer_len;     // Size of the row buffer
	uint64_t row_escape_count;   // Number of escape characters in the current row
	uint8_t  row_newline_length; // Length of the newline in the current row
	uint64_t row_file_offset;    // Offset of the current row in the file being processed
};

// Static function prototypes
static bool CsvReader_find_newline   (CsvReader *csv);
static void CsvReader_terminate_row  (CsvReader *csv);
static void CsvReader_advance_buffer (CsvReader *csv);

// Finds the next newline in the buffer and advances the buffer pointer to it's position
static bool CsvReader_find_newline (CsvReader *csv)
{
	while (csv->buffer_cur < csv->buffer_end)
	{
		char character = *csv->buffer_cur;

		if (character == csv->escape_char)
			csv->row_escape_count++;
		else if (character == '\n' && (csv->row_escape_count % 2) == 0)
			return true;

		csv->buffer_cur++;
	}

	return false;
}

// Replaces the first newline character with a null terminator
static void CsvReader_terminate_row (CsvReader *csv)
{
	if (csv->row_raw_len > 0 && csv->row[csv->row_raw_len - 1] == '\r')
	{
		csv->row_newline_length = 2;
		csv->row_raw_len--;
	}
	else
		csv->row_newline_length = 1;

	csv->row[csv->row_raw_len] = '\0';
}

// Reads the next 'CSV_BUFFER_SIZE' bytes from the file into the buffer
static void CsvReader_advance_buffer (CsvReader *csv)
{
	uint64_t size = fread(csv->buffer, sizeof(char), CSV_BUFFER_SIZE, csv->file);
	csv->buffer_end = csv->buffer + size;
	csv->buffer_cur = csv->buffer;
}

// Creates a new CSV reader with the default delimiter and escape characters
CsvReader *CsvReader_new (const char *filename)
{
	return CsvReader_new_full(filename, ',', '"');
}

// Creates a new CSV reader with the specified delimiter and escape characters
CsvReader *CsvReader_new_full (const char *filename, char delim_char, char escape_char)
{
	CsvReader *csv = malloc(sizeof(struct CsvReaderPrivate));

	// The file is null by default
	csv->file = NULL;

	// Open the CSV file for reading in binary mode
	if (filename && (csv->file = fopen(filename, "rb")) == NULL)
	{
		free(csv);
		return NULL;
	}

	// Allocating the row buffer and file buffer
	csv->buffer = csv->file ? malloc(CSV_BUFFER_SIZE) : NULL;
	csv->row = NULL;

	// Setting the delimiter and escape characters
	csv->delim_char = delim_char;
	csv->escape_char = escape_char;

	// Initializing the rest of the variables
	CsvReader_reset(csv);

	return csv;
}

// Resets the CSV reader to the beginning of the file
void CsvReader_reset (CsvReader *csv)
{
	// Resetting the file buffer
	if (csv->file) rewind(csv->file);
	uint64_t size = csv->file ? fread(csv->buffer, sizeof(char), CSV_BUFFER_SIZE, csv->file) : 0;

	// Resetting the buffer and column pointers
	csv->buffer_end = csv->buffer + size;
	csv->buffer_cur = csv->buffer;
	csv->column_cur = NULL;

	// Resetting the current column state
	csv->col_last_char = '\0';
	csv->col_raw_len = 0;
	csv->col_parse_len = 0;
	csv->col_row_offset = 0;

	// Resetting the current row state
	csv->row_raw_len = 0;
	csv->row_buffer_len = 0;
	csv->row_escape_count = 0;
	csv->row_newline_length = 0;
	csv->row_file_offset = 0;
}

// Deletes the CSV reader
void CsvReader_delete (CsvReader *csv)
{
	if (csv->file) fclose(csv->file);
	free(csv->buffer);
	free(csv->row);
	free(csv);
}

// Returns the next unparsed row in the CSV file
char *CsvReader_next_row (CsvReader *csv)
{
	// Increasing the file index
	csv->row_file_offset += csv->row_raw_len + csv->row_newline_length;

	// Resetting the column position
	csv->column_cur = NULL;

	// Resetting the current row state
	csv->row_raw_len = 0;
	csv->row_escape_count = 0;
	csv->row_newline_length = 0;

	// Scanning the file until a newline or EOF is found
	while (true)
	{
		// Finding the next newline
		char *row_start = csv->buffer_cur;
		bool newline_found = CsvReader_find_newline(csv);

		// Calculating the new length of the row
		uint64_t char_count = csv->buffer_cur - row_start;
		uint64_t new_row_length = csv->row_raw_len + char_count;

		// Increasing the row buffer size if necessary
		if (new_row_length > csv->row_buffer_len)
		{
			csv->row_buffer_len = new_row_length;
			csv->row = realloc(csv->row, new_row_length + 1);
		}

		// Copying the current row portion to the row buffer
		memcpy(csv->row + csv->row_raw_len, row_start, char_count);
		csv->row_raw_len = new_row_length;

		// If a newline was found, terminate the row and return it
		if (newline_found)
		{
			CsvReader_terminate_row(csv);
			csv->buffer_cur++;
			return csv->row;
		}
		// If the end of the file was reached, terminate the row and return it
		else if (feof(csv->file))
		{
			if (csv->row_raw_len > 0)
				CsvReader_terminate_row(csv);
			return (csv->row_raw_len > 0) ? csv->row : NULL;
		}
		else
		// If the end of the buffer was reached, advance the buffer
			CsvReader_advance_buffer(csv);
	}
}

// Returns the next parsed column in the current row
char *CsvReader_next_col (CsvReader *csv, char *row)
{
	// The row must be different from NULL
	if (row == NULL)
		return NULL;

	// If the 'column_cur' pointer is NULL, the first column is being read
	if (csv->column_cur == NULL)
	{
		csv->column_cur = row;
		csv->col_last_char = '\0';
		csv->col_row_offset = 0;
	}
	// Else the column offset is increased
	else
		csv->col_row_offset += csv->col_raw_len + 1;

	// Initializing the column pointers
	char *current_pos = csv->column_cur;
	char *col_start = csv->column_cur;
	char *col_end = csv->column_cur;

	// Initializing the column escaped state and length
	bool escaped = false;
	csv->col_raw_len = 0;
	csv->col_parse_len = 0;

	// If the current character is a delimiter, the column is empty
	if (*current_pos == csv->delim_char)
	{
		csv->col_last_char = *col_end;
		csv->column_cur++;
		*col_end = '\0';
		return col_start;
	}

	// If the current character is an escape character, the column is escaped
	if (*current_pos == csv->escape_char)
	{
		escaped = true;
		current_pos++;
		col_end++;
	}

	// Scanning the row until the end of the current column is found
	while (*current_pos)
	{
		// If the column is escaped, it is only terminated by a single matching escape character
		if (escaped)
		{
			if (*current_pos == csv->escape_char)
			{
				// If the escape character appears twice, it is escaped
				if (current_pos[1] == csv->escape_char)
					current_pos++;
				else
					break;
			}
		}
		// If the column is not escaped, it is always terminated by the delimiter character
		else if (*current_pos == csv->delim_char)
			break;

		// Copies the current character to the column end in case of an offset
		if (col_end != current_pos)
			*col_end = *current_pos;

		// The column pointers are advanced
		current_pos++;
		col_end++;
	}

	// If the current position is greater than the column start, the column is not empty
	if (current_pos > col_start)
	{
		// The position is increased to match the delimiter character or null terminator
		if (*current_pos && *current_pos != csv->delim_char)
			current_pos++;

		// If the last character isn't '\0' that means it is a delimiter, so the
		// column pointer is increased to match the next column start, also, the column
		// length is decreased as the delimiter character is not part of the column
		if (*current_pos)
		{
			current_pos++;
			csv->col_raw_len--;
		}

		// The column is terminated and it's state is updated
		csv->col_raw_len += current_pos - col_start;
		csv->col_parse_len = col_end - col_start;
		csv->column_cur = current_pos;
		csv->col_last_char = *col_end;
		*col_end = '\0';

		// The parsed length is adjusted if the column is escaped
		if (escaped) csv->col_parse_len--;

		return escaped ? col_start + 1 : col_start;
	}

	// If the last character read was a delimiter, the column is empty
	if (csv->col_last_char == csv->delim_char)
	{
		csv->col_last_char = '\0';
		return col_start;
	}

	csv->column_cur = NULL;
	return NULL;
}

// Returns the current row information
extern CsvRowInfo CsvReader_row_info (CsvReader *csv)
{
	return (CsvRowInfo)
	{
		.raw_len = csv->row_raw_len,
		.file_offset = csv->row_file_offset
	};
}

// Returns the current column information
extern CsvColInfo CsvReader_col_info (CsvReader *csv)
{
	return (CsvColInfo)
	{
		.raw_len = csv->col_raw_len,
		.parse_len = csv->col_parse_len,
		.row_offset = csv->col_row_offset
	};
}

#endif // PICO_CSV_IMPLEMENTATION