// Pico String Library
// by. Foxuper

#ifndef PICO_STRING_H
#define PICO_STRING_H

#include <stdint.h>

// String
typedef char *pstr;

// String Information
typedef struct pstri
{
	uint64_t length;
	uint64_t capacity;
}
pstri;

extern pstr  pstr_new     (char *string);
extern pstr  pstr_psnew   (pstr string);
extern pstr  pstr_new_exp (void *string, uint64_t length);
extern void  pstr_delete  (pstr string);
extern pstri pstr_info    (pstr string);
extern pstr  pstr_cpy     (pstr string, char *source);
extern pstr  pstr_pscpy   (pstr string, pstr source);
extern pstr  pstr_cpy_exp (pstr string, void *source, uint64_t source_len);
extern pstr  pstr_cat     (pstr string, char *source);
extern pstr  pstr_pscat   (pstr string, pstr source);
extern pstr  pstr_cat_exp (pstr string, void *source, uint64_t source_len);
extern int   pstr_cmp     (pstr string_a, char *string_b);
extern int   pstr_pscmp   (pstr string_a, pstr string_b);
extern int   pstr_cmp_exp (pstr string_a, void *string_b, uint64_t len_b);

#endif // PICO_STRING_H



#ifdef PICO_STRING_IMPLEMENTATION

// Private includes
#include <assert.h>
#include <stdlib.h>

// Header type bit definitions
#define TYPE_BITMASK  0x07
#define TYPE_BITCOUNT 3

// String type
enum pstrt
{
	PSTRH0,
	PSTRH8,
	PSTRH16,
	PSTRH32,
	PSTRH64,
	PSTRHNULL
};
typedef uint8_t pstrt;

// String headers (packed)
#pragma pack(push, 1)
	typedef struct pstrh0  { /* ----------------------------- */ pstrt type; } pstrh0;
	typedef struct pstrh8  { uint8_t  length; uint8_t  capacity; pstrt type; } pstrh8;
	typedef struct pstrh16 { uint16_t length; uint16_t capacity; pstrt type; } pstrh16;
	typedef struct pstrh32 { uint32_t length; uint32_t capacity; pstrt type; } pstrh32;
	typedef struct pstrh64 { uint64_t length; uint64_t capacity; pstrt type; } pstrh64;
#pragma pack(pop)

// Asserting string header sizes
static_assert(sizeof(pstrh0)  == 1,  "[pico_string] strheadt size is incorrect");
static_assert(sizeof(pstrh8)  == 3,  "[pico_string] strhead8 size is incorrect");
static_assert(sizeof(pstrh16) == 5,  "[pico_string] strhead16 size is incorrect");
static_assert(sizeof(pstrh32) == 9,  "[pico_string] strhead32 size is incorrect");
static_assert(sizeof(pstrh64) == 17, "[pico_string] strhead64 size is incorrect");

// Private functions
static void    *init_header   (void *memory, pstrt type, uint64_t length, uint64_t capacity);
static uint64_t header_size   (pstrt type);
static pstrt    required_type (uint64_t length);
static void     pstr_set_len  (pstr string, uint64_t length);
static pstr     increase_cap  (pstr string, uint64_t added_length);

// ---------------- Private definitions ------------------

#define PSTR_HEADER(T, string) \
	((pstrh ## T *) (string - sizeof(pstrh ## T)))

#define PSTR_TYPE(string) \
	(PSTR_HEADER(0, string)->type & TYPE_BITMASK)

#define GET_INFO_0 \
	uint64_t length = PSTR_HEADER(0, string)->type >> TYPE_BITCOUNT; \
	return (pstri) {length, length};

#define GET_INFO(T) \
	pstrh ## T *header = PSTR_HEADER(T, string); \
	return (pstri) {header->length, header->capacity}

#define HEADER_INIT_0 \
	*((pstrh0 *) memory) = (pstrh0) {capacity << TYPE_BITCOUNT | type}; \
	return (uint8_t *) memory + sizeof(pstrh0);

#define HEADER_INIT(T) \
	*((pstrh ## T *) memory) = (pstrh ## T) {length, capacity, type}; \
	return (uint8_t *) memory + sizeof(pstrh ## T)

// -------------------------------------------------------

pstr pstr_new (char *string)
{
	return pstr_new_exp(string, string ? strlen(string) : 0);
}

pstr pstr_psnew (pstr string)
{
	return pstr_new_exp(string, string ? pstr_info(string).length : 0);
}

pstr pstr_new_exp (void *string, uint64_t length)
{
	pstrt type = required_type(length);
	pstr new_string = malloc(header_size(type) + length + 1);

	if (new_string)
	{
		uint64_t capacity = length;
		length = string ? length : 0;
		new_string = init_header(new_string, type, length, capacity);

		if (string != NULL)
			memcpy(new_string, string, length);

		new_string[length] = '\0';
	}

	return new_string;
}

void pstr_delete (pstr string)
{
	if (string == NULL) return;
	uint64_t size = header_size(PSTR_TYPE(string));
	free(string - size);
}

pstri pstr_info (pstr string)
{
	switch (PSTR_TYPE(string))
	{
		case PSTRH0:  { GET_INFO_0;   }
		case PSTRH8:  { GET_INFO(8);  }
		case PSTRH16: { GET_INFO(16); }
		case PSTRH32: { GET_INFO(32); }
		case PSTRH64: { GET_INFO(64); }
	}
	return (pstri) {0, 0};
}

pstr pstr_cpy (pstr string, char *source)
{
	return pstr_cpy_exp(string, source, strlen(source));
}

pstr pstr_pscpy (pstr string, pstr source)
{
	return pstr_cpy_exp(string, source, pstr_info(source).length);
}

pstr pstr_cpy_exp (pstr string, void *source, uint64_t source_len)
{
	pstri info = pstr_info(string);

	if (info.capacity < source_len)
		string = increase_cap(string, source_len - info.length);

	if (string != NULL)
	{
		memcpy(string, source, source_len);
		pstr_set_len(string, source_len);
		string[source_len] = '\0';
	}

	return string;
}

pstr pstr_cat (pstr string, char *source)
{
	return pstr_cat_exp(string, source, strlen(source));
}

pstr pstr_pscat (pstr string, pstr source)
{
	return pstr_cat_exp(string, source, pstr_info(source).length);
}

pstr pstr_cat_exp (pstr string, void *source, uint64_t source_len)
{
	pstri info = pstr_info(string);
	string = increase_cap(string, source_len);

	if (string != NULL)
	{
		memcpy(string + info.length, source, source_len);
		pstr_set_len(string, info.length + source_len);
		string[info.length + source_len] = '\0';
	}

	return string;
}

int pstr_cmp (pstr string_a, char *string_b)
{
	return pstr_cmp_exp(string_a, string_b, strlen(string_b));
}

int pstr_pscmp (pstr string_a, pstr string_b)
{
	return pstr_cmp_exp(string_a, string_b, pstr_info(string_b).length);
}

int pstr_cmp_exp (pstr string_a, void *string_b, uint64_t len_b)
{
	uint64_t len_a = pstr_info(string_a).length;
	int comparison = memcmp(string_a, string_b, (len_a < len_b) ? len_a : len_b);
	return (comparison == 0) ? ((len_a > len_b) - (len_a < len_b)) : comparison;
}

static void *init_header (void *memory, pstrt type, uint64_t length, uint64_t capacity)
{
	switch (type)
	{
		case PSTRH0:  HEADER_INIT_0;
		case PSTRH8:  HEADER_INIT(8);
		case PSTRH16: HEADER_INIT(16);
		case PSTRH32: HEADER_INIT(32);
		case PSTRH64: HEADER_INIT(64);
	}
	return NULL;
}

static uint64_t header_size (pstrt type)
{
	switch (type & TYPE_BITMASK)
	{
		case PSTRH0:  return sizeof(pstrh0);
		case PSTRH8:  return sizeof(pstrh8);
		case PSTRH16: return sizeof(pstrh16);
		case PSTRH32: return sizeof(pstrh32);
		case PSTRH64: return sizeof(pstrh64);
	}
	return 0;
}

static pstrt required_type (uint64_t length)
{
	if (length <= 0x1F)       return PSTRH0;
	if (length <= 0xFF)       return PSTRH8;
	if (length <= 0xFFFF)     return PSTRH16;
	if (length <= 0xFFFFFFFF) return PSTRH32;
	return PSTRH64;
}

static void pstr_set_len (pstr string, uint64_t length)
{
	if (PSTR_TYPE(string) == PSTRH0)
		length = (length << TYPE_BITCOUNT) | PSTRH0;

	switch (PSTR_TYPE(string))
	{
		case PSTRH0:  PSTR_HEADER(0,  string)->type   = length; break;
		case PSTRH8:  PSTR_HEADER(8,  string)->length = length; break;
		case PSTRH16: PSTR_HEADER(16, string)->length = length; break;
		case PSTRH32: PSTR_HEADER(32, string)->length = length; break;
		case PSTRH64: PSTR_HEADER(64, string)->length = length; break;
	}
}

static pstr increase_cap (pstr string, uint64_t added_length)
{
	pstri info = pstr_info(string);
	uint64_t new_capacity = info.length + added_length;

	pstrt old_type = PSTR_TYPE(string);
	pstrt new_type = required_type(new_capacity);

	if (new_capacity > info.capacity)
	{
		if (old_type == new_type)
		{
			uint64_t size = header_size(old_type);
			string = realloc(string - size, size + new_capacity + 1);
			return init_header(string, new_type, info.length, new_capacity);
		}
		else
		{
			pstr new_string = pstr_new_exp(NULL, new_capacity);

			if (new_string)
			{
				memcpy(new_string, string, info.length);
				pstr_set_len(new_string, info.length);
				new_string[info.length] = '\0';
			}

			pstr_delete(string);
			return new_string;
		}
	}

	return string;
}

#endif // PICO_STRING_IMPLEMENTATION